#ifndef MY_UTILS_H__
#define MY_UTILS_H__

#include "adapter.h"
#include "specializations.h"

/** Usage:
 *  char bigString[1000];
 *  char *p = bigString;
 *  bigString[0] = '\0';
 *  p = xstrcat(p,"John, ");
 *  p = xstrcat(p,"Paul, ");
 *  p = xstrcat(p,"George, ");
 *  p = xstrcat(p,"Joel ");
 * This method is a fast way to concatenate strings. When the method completes, it has a pointer
 * to the end of current string in the buffer where the next concatenation will take place. This method
 * is good when you need to create a long string by adding several small strings to it. It avoids the
 * search to the end of the string that the standard
 * @param dest: buffer that the string will be concatenated to
 * @param src: the string to be concatenated
 * @return a pointer to the end of the string in the buffer where the next concatenation will take place.
 */
char* xstrcat(char* dest, const char* src);

char* byteToHex(unsigned char* bytes, char* buffer, char* separator, unsigned long count);

/**@brief Function for initializing services that will be used by the application.
 *
 * Create the specified primary standard BTLE service and add it to GATT
 *
 * @param [IN] the adapter for sending over the serial port. Needed in all sd_* calls using the driver.
 * @param [IN] serviceHandle pointer to a variable to take the created service's handle
 * @param the 16-bit service value, so it needs to be a BTLE standardized service
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
unsigned long createPrimaryService(adapter_t *adapter, unsigned short* serviceHandle, unsigned short serviceUuid);

/**@brief Function for adding the a standard BTLE characteristic.
 *
 * @param adapter: the serial port adapter that all sd_* calls have when using the pc-ble-driver
 * @param serviceHandle: the handle of the service to which the characteristic belongs
 * @param charHandles: a pointer to a struct that takes the handles of the created characteristic (declaration, value, and descriptor - if present)
 * @param uuid: the 16-bit UUID of the characteristic
 * @param hasCCCD: true if the characteristic is to include a CCCD
 * @param indicate: true if indicate, otherwise notify. Ignored if hasCCCD is false
 * @param cccdWritePermission: the security level of the CCCD write operation (ignored if no CCCD)
 * @param attrReadPermission: the security level of the characteristic value read operation. If security mode is 0 no read at all.
 * @param attrWritePermission: the security level of the characteristic value write operation. If security mode is 0 no write at all.
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
unsigned long createStandardCharacteristic(adapter_t* adapter,
	uint16_t serviceHandle,
	ble_gatts_char_handles_t* charHandles,
	uint16_t uuid,
	bool hasCCCD,
	bool indicate,
	ble_gap_conn_sec_mode_t cccdWritePermission,
	ble_gap_conn_sec_mode_t attrReadPermission,
	ble_gap_conn_sec_mode_t attrWritePermission);

/**
 * Method to generate a Device Information Service
 * @param adapter: the serial port adapter that all sd_* calls have when using the pc-ble-driver
 * @param [out]serviceHandle: a pointer to the service handle for the Device Information Service.
 * @param disCharValues: pointer to a struct containing the characteristic values in the service. If NULL that characteristic
 *        will not be created.
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
unsigned long createDeviceInformationService(adapter_t* adapter, short int* serviceHandle, s_disCharValues* disCharValues);

/**
 * Method generates advertisement data. Several defaults are assumed
 * @param adapter: the serial port adapter that all sd_* calls have when using the pc-ble-driver
 * @param advHandle Advertising handle to advertise on, received from @ref sd_ble_gap_adv_set_configure. Do not fully understand this
 * @param list of 16-bit service UUIDs
 * @param length the number of service UUIDs
 * @param 16-byte array of a 128-bit UUID in littleendian order
 * @param the friendly name of the device
 */
uint32_t advertisement_data_set(adapter_t* adapter, uint8_t* advHandle, short int* serviceUuids, int length,
	uint8_t* service128Uuids, char* friendlyName);

/**
 * Method saves bonding data to a file
 * @param fileName name of the file where the bonding data is to be stored
 * @param keys the peripheral and central keys (LTK, IRK, and others)
 * @param saveDataBuffer a buffer to hold the data obtained from the sd_ble_gatts_sys_attr_get() method
 * @param saveDataLength the length of the data obtained from the sd_ble_gatts_sys_attr_get()
 * @param cccdSet a pointer to the enabled state of the characteristics
 */
void saveKeysToFile(char* fileName, ble_gap_sec_keyset_t* keys, 
    unsigned char* saveDataBuffer, unsigned short int* saveDataLength, bool* cccdSet);

/**
 * Method loads saved bonding data from a file
 * @param fileName name of the file where the bonding data is stored
 * @param keys the peripheral and central keys (LTK, IRK, and others)
 * @param saveDataBuffer a buffer to hold the data obtained from the sd_ble_gatts_sys_attr_get() method
 * @param saveDataLength the length of the data obtained from the sd_ble_gatts_sys_attr_get()
 * @param cccdState a pointer to the enabled state of the characteristics
 */
void loadKeysFromFile(char* fileName, ble_gap_sec_keyset_t* keys, 
    unsigned char* saveDataBuffer, unsigned short int* saveDataLength, bool* cccdSet);

#endif // MY_UTILS_H__