/* Utilities.c */

#include <stdbool.h>
#include <stdio.h>
#include <malloc.h>
#include <stddef.h>
#include <string.h>
#include "ble.h"
#include "ble_types.h"
#include "BTLE_Assigned_Numbers.h"
#include "DISControlStructs.h"

#define OPCODE_LENGTH 1                                                 /**< Length of opcode inside Heart Rate Measurement packet. */
#define HANDLE_LENGTH 2                                                 /**< Length of handle inside Heart Rate Measurement packet. */

#define MAX_HRM_LEN (BLE_EVT_LEN_MAX(GATT_MTU_SIZE_DEFAULT) - OPCODE_LENGTH - HANDLE_LENGTH)
#ifndef GATT_MTU_SIZE_DEFAULT
#define GATT_MTU_SIZE_DEFAULT BLE_GATT_ATT_MTU_DEFAULT
#endif

#pragma region // -------------- xstrcat
/* Usage:
 *  char bigString[1000];
 *  char *p = bigString;
 *  bigString[0] = '\0';
 *  p = xstrcat(p,"John, ");
 *  p = xstrcat(p,"Paul, ");
 *  p = xstrcat(p,"George, ");
 *  p = xstrcat(p,"Joel ");
 */
char* xstrcat(char* dest, const char* src)
{
    while (*dest) dest++;
    while (*dest++ = *src++);
    return --dest;
}
#pragma endregion

#pragma region // -------------- byteToHex
char* byteToHex(unsigned char* bytes, char* buffer, char* separator, unsigned long count)
{
    char* cptr = buffer;
    unsigned long i;
    char buf[4];
    *cptr = 0;
    for (i = 0; i < count - 1; i++)
    {
        sprintf(buf, "%02X", bytes[i]);
        cptr = xstrcat(cptr, buf);
        cptr = xstrcat(cptr, separator);
    }
    sprintf(buf, "%02X", bytes[i]);
    cptr = xstrcat(cptr, buf);
    return buffer;
}
#pragma endregion

#pragma region // -------------- createPrimaryService
unsigned long createPrimaryService(adapter_t *adapter, unsigned short* serviceHandle, unsigned short serviceUuid)
{
    unsigned long    error_code;
    ble_uuid_t  ble_uuid;

    BLE_UUID_BLE_ASSIGN(ble_uuid, serviceUuid);

    error_code = sd_ble_gatts_service_add(adapter,
        BLE_GATTS_SRVC_TYPE_PRIMARY,
        &ble_uuid,
        serviceHandle);

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to initialize service. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }
    printf("Service initiated\n");
    fflush(stdout);
    return NRF_SUCCESS;
}
#pragma endregion

#define OPCODE_LENGTH 1                                                 /**< Length of opcode inside packet. */
#define HANDLE_LENGTH 2                                                 /**< Length of handle inside packet. */


#pragma region // -------------- createStandardCharacteristic
unsigned long createStandardCharacteristic(adapter_t* adapter,
    uint16_t serviceHandle,
    ble_gatts_char_handles_t* charHandles,
    uint16_t uuid,
    bool hasCCCD,
    bool indicate,
    ble_gap_conn_sec_mode_t cccdWritePermission,
    ble_gap_conn_sec_mode_t attrReadPermission,
    ble_gap_conn_sec_mode_t attrWritePermission)
{
    unsigned long       error_code;
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_md_t cccd_md;
    ble_gatts_attr_md_t* cccd_md_ptr = NULL;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;
    uint8_t             encoded_initial_value[MAX_HRM_LEN];

    if (hasCCCD)
    {
        memset(&cccd_md, 0, sizeof(cccd_md));

        cccd_md.rd_auth = 1;    // Added to trap a read of the descriptor
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
        cccd_md.write_perm.lv = cccdWritePermission.lv;
        cccd_md.write_perm.sm = cccdWritePermission.sm;
        cccd_md.vloc = BLE_GATTS_VLOC_STACK;
        cccd_md_ptr = &cccd_md;
    }

    memset(&char_md, 0, sizeof(char_md));

    if (hasCCCD)
    {
        if (indicate)
        {
            char_md.char_props.indicate = 1;
        }
        else
        {
            char_md.char_props.notify = 1;
        }
    }
    char_md.char_props.read = ((attrReadPermission.sm == 0) ? 0 : 1);
    char_md.char_props.write = ((attrWritePermission.sm == 0) ? 0 : 1);
    char_md.p_char_user_desc = NULL;
    char_md.p_char_pf = NULL;
    char_md.p_user_desc_md = NULL;
    char_md.p_cccd_md = cccd_md_ptr;
    char_md.p_sccd_md = NULL;

    BLE_UUID_BLE_ASSIGN(ble_uuid, uuid);

    memset(&attr_md, 0, sizeof(attr_md));

    attr_md.read_perm.sm = attrReadPermission.sm;
    attr_md.read_perm.lv = attrReadPermission.lv;
    attr_md.write_perm.sm = attrWritePermission.sm;
    attr_md.write_perm.lv = attrWritePermission.lv;

    attr_md.vloc = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 1;
    attr_md.wr_auth = 0;
    attr_md.vlen = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len = 1; //attr_char_value_init_len;
    attr_char_value.init_offs = 0;
    attr_char_value.max_len = MAX_HRM_LEN;
    attr_char_value.p_value = encoded_initial_value;

    error_code = sd_ble_gatts_characteristic_add(adapter, serviceHandle,
        &char_md,
        &attr_char_value,
        charHandles);

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to initialize characteristics. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }

    printf("Characteristic initiated\n");
    fflush(stdout);
    return NRF_SUCCESS;
}
#pragma endregion

#pragma region // -------------- createDeviceInformationService
static ble_gatts_char_handles_t m_systenId_handle;
static ble_gatts_char_handles_t m_manufacturer_handle;
static ble_gatts_char_handles_t m_model_handle;
static ble_gatts_char_handles_t m_serial_handle;
static ble_gatts_char_handles_t m_firmware_handle;
static ble_gatts_char_handles_t m_software_handle;
static ble_gatts_char_handles_t m_hardware_handle;
static ble_gatts_char_handles_t m_regcert_handle;
unsigned long createDeviceInformationService(adapter_t* adapter, short int *serviceHandle, s_disCharValues *disCharValues)
{
    uint32_t error_code = createPrimaryService(adapter, serviceHandle, BTLE_DEVICE_INFORMATION_SERVICE);
    if (error_code != NRF_SUCCESS)
    {
        printf("Unable to create Device Information Service. Error code 0x%X\n", error_code);
        return error_code;
    }

    if(disCharValues->systemId != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
            &m_systenId_handle,
            BTLE_DIS_SYSTEM_ID_CHAR,
            false,
            false,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create System Id characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->manufacturerName != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
            &m_manufacturer_handle,
            BTLE_DIS_MANUFACTURER_DEVICE_NAME_CHAR,
            false,
            false,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create Manufacturer Name characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->modelNumber != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
            &m_model_handle,
            BTLE_DIS_MODEL_NUMBER_CHAR,
            false,
            false,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create Model Number characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->serialNumber != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
            &m_serial_handle,
            BTLE_DIS_SERIAL_NUMBER_CHAR,
            false,
            false,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create Serial Number characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->firmwareRevision != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
            &m_firmware_handle,
            BTLE_DIS_FIRMWARE_REVISION_CHAR,
            false,
            false,
            (ble_gap_conn_sec_mode_t) {    0, 0},
            (ble_gap_conn_sec_mode_t) {    1, 1},
            (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create Firmware Revision characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->hardwareRevision != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
        &m_hardware_handle,
        BTLE_DIS_HARDWARE_REVISION_CHAR,
        false,
        false,
        (ble_gap_conn_sec_mode_t) {    0, 0},
        (ble_gap_conn_sec_mode_t) {    1, 1},
        (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create Hardware Revision characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->softwareRevision != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
        &m_software_handle,
        BTLE_DIS_SOFTWARE_REVISION_CHAR,
        false,
        false,
        (ble_gap_conn_sec_mode_t) {    0, 0},
        (ble_gap_conn_sec_mode_t) {    1, 1},
        (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create Software Revision characteristic. Error code 0x%X\n", error_code);
        }
    }
    if (disCharValues->regCertDataList != NULL)
    {
        error_code = createStandardCharacteristic(adapter, *serviceHandle,
        &m_regcert_handle,
        BTLE_DIS_REG_CERT_DATA_LIST_CHAR,
        false,
        false,
        (ble_gap_conn_sec_mode_t) {    0, 0},
        (ble_gap_conn_sec_mode_t) {    1, 1},
        (ble_gap_conn_sec_mode_t) {    0, 0});

        if (error_code != NRF_SUCCESS)
        {
            printf("Unable to create RegCertDataList characteristic. Error code 0x%X\n", error_code);
        }
    }
    return NRF_SUCCESS;
}
#pragma endregion

#define ADVERTISING_INTERVAL_40_MS 64  /**< 0.625 ms = 40 ms */
#define BUFFER_SIZE 30
static ble_gap_adv_params_t     m_adv_params;
uint32_t advertisement_data_set(adapter_t* adapter, uint8_t* advHandle, short int* serviceUuids, int length, 
    uint8_t* service128Uuids, char* friendlyName)
{
    uint32_t error_code;
    uint8_t  index = 0;
    uint8_t  scan_index = 0;
    uint8_t  i = 0;
    uint8_t  data_buffer[BUFFER_SIZE];
    uint8_t  scan_rsp_buffer[BUFFER_SIZE];

    const uint8_t name_length = (uint8_t)strlen(friendlyName);

    // Set the flags - important to indicate no EDR/BR support for some Androids
    // Actually by spec one is SUPPOSED to set one of the discoverable flags to be discoverable.
    data_buffer[index++] = 2;        // length of flags data type
    data_buffer[index++] = BLE_GAP_AD_TYPE_FLAGS;
    data_buffer[index++] = BLE_GAP_ADV_FLAG_LE_GENERAL_DISC_MODE | BLE_GAP_ADV_FLAG_BR_EDR_NOT_SUPPORTED;    // No EDR/BR support and general discoverable

    // Set the device name.
    if (friendlyName != NULL)
    {
        data_buffer[index++] = name_length + 1; // Device name + data type
        data_buffer[index++] = BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME;
        memcpy((char*)&data_buffer[index], friendlyName, name_length);
        index += name_length;
    }

    if (serviceUuids != NULL)
    {
        // Set the device's available services.
        data_buffer[index++] = 1 + 2 * length;
        data_buffer[index++] = BLE_GAP_AD_TYPE_16BIT_SERVICE_UUID_COMPLETE;
        for (i = 0; i < length; i++)
        {
            // Store in little-endian format.
            data_buffer[index++] = serviceUuids[i] & 0xFF;
            data_buffer[index++] = (serviceUuids[i] & 0xFF00) >> 8;
        }
    }

    // Scan response:
    if (service128Uuids != NULL)
    {
        scan_rsp_buffer[scan_index++] = 1 + 16; //(only room for one)
        scan_rsp_buffer[scan_index++] = BLE_GAP_AD_TYPE_128BIT_SERVICE_UUID_MORE_AVAILABLE;
        // Store in little-endian format.
        memcpy(&scan_rsp_buffer[scan_index], service128Uuids, 16);
        scan_index = scan_index + 16;
    }


    ble_gap_adv_properties_t adv_properties;
    adv_properties.type = BLE_GAP_ADV_TYPE_CONNECTABLE_SCANNABLE_UNDIRECTED;
    adv_properties.anonymous = 0;
    adv_properties.include_tx_power = 0;

    m_adv_params.properties = adv_properties;
    m_adv_params.filter_policy = BLE_GAP_ADV_FP_ANY;
    m_adv_params.duration = BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED;
    m_adv_params.p_peer_addr = NULL;
    m_adv_params.interval = ADVERTISING_INTERVAL_40_MS;
    m_adv_params.max_adv_evts = 0;
    m_adv_params.primary_phy = BLE_GAP_PHY_AUTO;
    m_adv_params.secondary_phy = BLE_GAP_PHY_AUTO;
    m_adv_params.channel_mask[0] = 0;
    m_adv_params.channel_mask[1] = 0;
    m_adv_params.channel_mask[2] = 0;
    m_adv_params.channel_mask[3] = 0;
    m_adv_params.channel_mask[4] = 0;

    ble_gap_adv_data_t m_adv_data;

    ble_data_t adv_data;
    adv_data.p_data = data_buffer;
    adv_data.len = index;

    ble_data_t scan_rsp_data;
    memset(&scan_rsp_buffer, 0, sizeof(scan_rsp_buffer));
    scan_rsp_data.p_data = NULL;
    scan_rsp_data.len = 0;
    if (scan_index > 0)
    {
        scan_rsp_data.p_data = scan_rsp_buffer;
        scan_rsp_data.len = scan_index;
    }
    m_adv_data.scan_rsp_data = scan_rsp_data;

    m_adv_data.adv_data = adv_data;

    error_code = sd_ble_gap_adv_set_configure(adapter, advHandle, &m_adv_data, &m_adv_params);

    if (error_code != NRF_SUCCESS)
    {
        printf("Failed to set advertisement data. Error code: 0x%02X\n", error_code);
        fflush(stdout);
        return error_code;
    }

    printf("Advertising data set\n");
    fflush(stdout);
    return NRF_SUCCESS;
}

#pragma region // -------------- saveKeysToFile
void saveKeysToFile(char* fileName, ble_gap_sec_keyset_t* keys,
    unsigned char *saveDataBuffer, unsigned short int *saveDataLength, bool* cccdSet)
{
    FILE* f = fopen(fileName, "wb");  // create and/or overwrite
    if (!f)
    {
        int error;
        _get_errno(&error);
        printf("Error in opening file %s for saving data. Error: %d Aborting.\n", fileName, error);
        return;
    }

    // write one own LTK to the file *f
    fwrite(keys->keys_own.p_enc_key, sizeof(ble_gap_enc_key_t), 1, f);
    fwrite(keys->keys_own.p_id_key, sizeof(ble_gap_id_key_t), 1, f);
    fwrite(keys->keys_peer.p_enc_key, sizeof(ble_gap_enc_key_t), 1, f);
    fwrite(keys->keys_peer.p_id_key, sizeof(ble_gap_id_key_t), 1, f);
    fwrite(cccdSet, sizeof(bool), 1, f);
    fwrite(saveDataLength, sizeof(unsigned short int), 1, f);
    if (*saveDataLength > 0 && saveDataBuffer != NULL)
    {
        fwrite(saveDataBuffer, *saveDataLength * sizeof(unsigned char), 1, f);
    }

    fclose(f);
}
#pragma endregion

#pragma region // -------------- loadKeysFromFile
void loadKeysFromFile(char* fileName, ble_gap_sec_keyset_t* keys, 
    unsigned char* saveDataBuffer, unsigned short int *saveDataLength, bool *cccdSet)
{
    FILE* f = fopen(fileName, "rb");  // create and/or overwrite
    if (!f)
    {
        int error;
        _get_errno(&error);
        printf("Error in creating or opening file %s for reading. Error: %d Aborting.\n", fileName, error);
        return;
    }

    // write one own LTK to the file *f
    fread(keys->keys_own.p_enc_key, sizeof(ble_gap_enc_key_t), 1, f);
    fread(keys->keys_own.p_id_key, sizeof(ble_gap_id_key_t), 1, f);
    fread(keys->keys_peer.p_enc_key, sizeof(ble_gap_enc_key_t), 1, f);
    fread(keys->keys_peer.p_id_key, sizeof(ble_gap_id_key_t), 1, f);
    fread(cccdSet, sizeof(bool), 1, f);
    fread(saveDataLength, sizeof(unsigned short), 1, f);
    if (saveDataLength > 0)
    {
        saveDataBuffer = calloc(1, (*saveDataLength) * sizeof(unsigned char));
        fwrite(saveDataBuffer, *saveDataLength * sizeof(unsigned char), 1, f);
    }

    fclose(f);
}
#pragma endregion